// MZSplit - a small and simple file splitter
//
// Copyright 2005-2007 Marcello Zaniboni
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 2.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <cstring>
#include <cctype>
#include <cstdlib>
#include <cstdio>
#include <FL/Fl_Window.H>
#include <FL/Fl_Widget.H>
#include <FL/Fl_Input.H>
#include <FL/Fl_Int_Input.H>
#include <FL/Fl_Round_Button.H>
#include <FL/Fl_Check_Button.H>
#include <FL/Fl_File_Chooser.H>
#include <FL/Fl_Progress.H>
#include <FL/Fl_Help_View.H>
#include <FL/Enumerations.H>
#include <FL/fl_draw.H>
#include <FL/fl_ask.H>
#include <FL/filename.H>
#include <FL/x.H>

#include "msgbox.h"
#include "callbacks.h"
#include "processing.h"
#include "helpdocuments.h"

// file-scope prototypes
void create_help_window();

static const int ONE_K = 1024;
static const long long SIZES_IN_BYTES[] = {
    1457000,               // floppy
    650   * ONE_K * ONE_K, // CD-ROM
    ONE_K * ONE_K          // MiB
};

static long long current_size;

static Fl_Window *main_win = NULL;
static Fl_Round_Button *floppy = NULL;
static Fl_Round_Button *cdrom = NULL;
static Fl_Round_Button *custom = NULL;
static Fl_Int_Input *custom_size = NULL;
static Fl_Input *in_file = NULL;
static Fl_Input *out_dir = NULL;
static Fl_Button *out_dir_browse = NULL;
static Fl_Window *progress_win = NULL;
static Fl_Input *progress_status = NULL;
static Fl_Progress *progress_bar = NULL;
static Fl_Window *help_win = NULL;
static Fl_Help_View* help_document = NULL;

void init_references(
        Fl_Window *w1,
        Fl_Round_Button *r1,
        Fl_Round_Button *r2,
        Fl_Round_Button *r3,
        Fl_Int_Input *s,
        Fl_Input *i1,
        Fl_Input *i2,
        Fl_Button *b1,
        Fl_Window *w2,
        Fl_Input *o1,
        Fl_Progress *pb
        ) {
    main_win = w1;
    floppy = r1;
    cdrom = r2;
    custom = r3;
    custom_size = s;
    in_file = i1;
    out_dir = i2;
    out_dir_browse = b1;
    progress_win = w2;
    progress_status = o1;
    progress_bar = pb;
}

void cb_no_esc_key_window(Fl_Widget *o, void *v) {
    if (Fl::event() == FL_SHORTCUT && Fl::event_key() == FL_Escape) return;
    exit(0);
}

void cb_radio_behaviour(Fl_Widget *o , void *v) {
    if (custom->value())
        custom_size->activate();
    else
        custom_size->deactivate();
}

void cb_default_folder_behaviour(Fl_Widget *o , void *v) {
    if(((Fl_Check_Button *) v)->value()) {
        out_dir->deactivate();
        out_dir->value("");
        out_dir_browse->deactivate();
    } else {
        out_dir->activate();
        out_dir_browse->activate();
    }
}

void cb_custom_size_behaviour(Fl_Widget *o , void *v) {
    const int BUFFSIZE = 512;
    char new_value[BUFFSIZE];
    char value_copy[BUFFSIZE];
    for (int i = 0; i < BUFFSIZE; i++) value_copy[i] = new_value[i] = '\0';
    strncpy(value_copy, custom_size->value(), BUFFSIZE - 1);
    for (int i = 0, j = 0; i < strlen(value_copy); i++)
        if (isdigit(value_copy[i]))
            new_value[j++] = value_copy[i];
    if (strlen(new_value) > 4) {
        new_value[4] = '\0';
        display_warning_message(fl_xid(main_win), "Maximum 4 digits are allowed in\nthe custom size input box.\nThe value has been changed.");
    }
    custom_size->value(new_value);
}

void cb_file_browsing(Fl_Widget *o , void *filename_box) {
    char *fname = fl_file_chooser(
        "MZSplit - Select the file to split / join",
        "*", "\\", 0);
    if (fname != NULL) {
        for (int i = 0; i < strlen(fname); i++)
            if (fname[i] == '/') fname[i] = '\\';
        ((Fl_Input *) filename_box)->value(fname);
        ((Fl_Input *) filename_box)->position(
            strlen(((Fl_Input *) filename_box)->value()),
            strlen(((Fl_Input *) filename_box)->value()));
    }
}

void cb_folder_browsing(Fl_Widget *o , void *foldername_box) {
    char *fname = fl_dir_chooser(
        "MZSplit - Select the output folder", "\\", 0);
    if (fname != NULL) {
        for (int i = 0; i < strlen(fname); i++)
            if (fname[i] == '/') fname[i] = '\\';
        ((Fl_Input *) foldername_box)->value(fname);
        ((Fl_Input *) foldername_box)->position(
            strlen(((Fl_Input *) foldername_box)->value()),
            strlen(((Fl_Input *) foldername_box)->value()));
    }
}

void cb_help(Fl_Widget *o , void *foldername_box) {
    if (!help_win) create_help_window();
    help_document->value(helpstring);
    help_document->textsize(16);
    help_win->show();
}

void cb_about(Fl_Widget *o , void *foldername_box) {
    if (!help_win) create_help_window();
    help_document->value(gnugplstring);
    help_document->textsize(11);
    help_win->show();
}

void cb_go(Fl_Widget *o , void *v) {
    // user input validation
    char canonical_foldername[FL_PATH_MAX + 1];
    FILE *testfile = fopen(in_file->value(), "rb");
    if (!testfile) {
        display_error_message(fl_xid(main_win), "The file does not exist!");
        return;
    }
    fclose(testfile);
    if(strlen(out_dir->value())) {
        strncpy(canonical_foldername, out_dir->value(), FL_PATH_MAX);
        canonical_foldername[strlen(out_dir->value())] = '\0';
        for (int i = 0; i < strlen(canonical_foldername); i++)
            if (canonical_foldername[i] == '/') canonical_foldername[i] = '\\';
        while (strlen(canonical_foldername) > 1 && canonical_foldername[strlen(canonical_foldername) - 1] == '\\')
            canonical_foldername[strlen(canonical_foldername) - 1] = '\0';
        if (!fl_filename_isdir(canonical_foldername)) {
            display_error_message(fl_xid(main_win), "The output folder does not exist!");
            return;
        }
    } else {
        canonical_foldername[0] = '\0';
    }

    // size computing
    if (floppy->value())
        current_size = SIZES_IN_BYTES[0];
    else if (cdrom->value())
        current_size = SIZES_IN_BYTES[1];
    else if (custom->value())
        current_size = SIZES_IN_BYTES[2] * atol(custom_size->value());

    // data processing
    main_win->hide();
    progress_win->show();
    progress_win->cursor(FL_CURSOR_WAIT);
    process(
        in_file->value(),
        current_size,
        canonical_foldername,
        progress_win,
        progress_status,
        progress_bar);
    main_win->show();
    progress_win->hide();
}

void create_help_window() {
    help_win = new Fl_Window(710, 445, "MZSplit - help window");
    help_win->labelsize(12);
    help_document = new Fl_Help_View(5, 5, 700, 435);
    help_document->textfont(FL_HELVETICA);
    help_document->textcolor(FL_DARK_BLUE);
    help_document->color(
        fl_color_cube(
            255 * (FL_NUM_RED   - 1) / 255,
            255 * (FL_NUM_GREEN - 1) / 255,
            216 * (FL_NUM_BLUE  - 1) / 255)
    );
    help_win->end();
}
