// MZSplit - a small and simple file splitter
//
// Copyright 2005-2007 Marcello Zaniboni
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 2.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cctype>
#include <ctime>
#include <windows.h>
#include <sys/stat.h>

#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Input.H>
#include <FL/Fl_Progress.H>
#include <FL/fl_ask.H>
#include <FL/filename.H>
#include <FL/x.H>
#include <FL/Enumerations.H>

#include "processing.h"
#include "msgbox.h"

static const int ONE_K = 1024;

void process(
        const char *file_name,
        long long file_size,
        const char *out_folder,
        Fl_Window *progress_window,
        Fl_Input *progress_status,
        Fl_Progress *progress_bar) {

    progress_status->value(fl_filename_name(file_name));
    progress_status->position(strlen(progress_status->value()), strlen(progress_status->value()));

    // file names: init
    int fnamelen = strlen(file_name);
    int join = 0; // 0 = split; 1 = join
    if (tolower(file_name[fnamelen - 1]) == 'a' &&
        tolower(file_name[fnamelen - 2]) == 'a' &&
        tolower(file_name[fnamelen - 3]) == '-' &&
        tolower(file_name[fnamelen - 4]) == 's' &&
        tolower(file_name[fnamelen - 5]) == '.') join = 1;
    char currext[] = "-aa";
    char outputfilename[FL_PATH_MAX + 1];

    // file names: determination
    if (strlen(out_folder)) {
        strncpy(outputfilename, out_folder, FL_PATH_MAX);
        outputfilename[strlen(out_folder)] = '\0';
        strcat(outputfilename, "\\");
        strncat(outputfilename, fl_filename_name(file_name), FL_PATH_MAX);
    } else {
        outputfilename[0] = '\0';
        strncat(outputfilename, file_name, FL_PATH_MAX);
    }
    if (join) {
        outputfilename[strlen(outputfilename) - 5] = '\0';
    } else {
        strcat(outputfilename, ".s");
        strcat(outputfilename, currext);
    }

    // file processing
    FILE *fin, *fout;
    if (join) {
        long long sourcefilelength = readfilelength(file_name);
        long long bytecounter = 0;
        int erroroccurred = 0;
        int endofjoin = 0;
        if ((fout = fopen(outputfilename, "wb")) != NULL) {
            if ((fin = fopen(file_name, "rb")) != NULL) {
                char inputbuffer[ONE_K];
                while (!(endofjoin || erroroccurred)) {
                    if (feof(fin)) {
                        fclose(fin);
                        nextsplitextension(currext);
                        int tmp_l = strlen(file_name);
                        char file_name_new[FL_PATH_MAX + 1];
                        strcpy(file_name_new, file_name);
                        for (int i = 1; i < 6; i++) file_name_new[tmp_l - i] = '\0';
                        strcat(file_name_new, ".s");
                        strcat(file_name_new, currext);
                        sourcefilelength = readfilelength(file_name_new);
                        bytecounter = 0;
                        progress_bar->label("100 %");
                        progress_bar->value((double)100);
                        Fl::check();
                        Sleep(120);
                        if ((fin = fopen(file_name_new, "rb")) == NULL) {
                            endofjoin = 1;
                        } else {
                            progress_bar->label("0 %");
                            progress_bar->value((double)0);
                            progress_status->value(fl_filename_name(file_name_new));
                            progress_status->position(strlen(progress_status->value()), strlen(progress_status->value()));
                            Fl::check();
                            Sleep(120);
                        }
                    }
                    if (!endofjoin) {
                        int n_read = fread(inputbuffer, sizeof(char), ONE_K, fin);
                        bytecounter += (long long)n_read;
                        if (!(bytecounter % (16 * ONE_K))) {
                            if (refresh()) {
                                char lbl_val[50];
                                double p = ((double)bytecounter / (double)sourcefilelength) * (double)100;
                                sprintf(lbl_val, "%d %%\0", (int)p);
                                progress_bar->value(p);
                                progress_bar->label(lbl_val);
                                progress_window->cursor(FL_CURSOR_WAIT);
                                Fl::check();
                            }
                        }
                        int n_written = fwrite(inputbuffer, sizeof(char), n_read, fout);
                        if (n_written != n_read) {
                            erroroccurred = 1;
                            fclose(fout);
                            display_error_message(fl_xid(progress_window), "ERROR writing output file!!!");
                        }
                    }
                    // closed window check (forced exit)
                    if (!progress_window->shown()) {
                        progress_window->show();
                        display_warning_message(
                            fl_xid(progress_window),
                            "The joining process has been interrupted!\nThe output file is incomplete; please\nremove it manually.");
                        fclose(fout);
                        fclose(fin);
                        return;
                    }
                }
                progress_bar->label("please wait a little more...");
                progress_window->cursor(FL_CURSOR_WAIT);
                Fl::check();
                if (!(endofjoin || erroroccurred)) fclose(fin);
            } else {
                display_error_message(fl_xid(progress_window), "Cannot open input file!");
            }
            fclose(fout);
            if (endofjoin && (!erroroccurred)) {
                progress_bar->label("done !");
                Fl::check();
                char *endmsg = (char *)calloc(512, sizeof(char));
                strcpy(endmsg, "Input files successfully joined! Output file name:\n");
                strcat(endmsg, fl_filename_name(outputfilename));
                display_message(fl_xid(progress_window), endmsg);
                free(endmsg);
            }
        } else {
            display_error_message(fl_xid(progress_window), "Cannot open input file!");
        }
    } else {
        const long long SOURCEFILELENGTH = readfilelength(file_name);
        const long long OUTFILESIZELIMIT = file_size;
        progress_status->value(fl_filename_name(file_name));
        if (SOURCEFILELENGTH <= 0 || SOURCEFILELENGTH <= OUTFILESIZELIMIT) {
            display_warning_message(fl_xid(progress_window), "Nothing to split: the selected size is greater than the file size.");
        } else {
            long long bytecounter = 0;
            if ((fin = fopen(file_name, "rb")) != NULL) {
                if ((fout = fopen(outputfilename, "wb")) != NULL) {
                    long long outputbytecounter = 0;
                    int erroroccurred = 0;
                    int changeoutputfile = 0;
                    char inputbuffer[ONE_K];
                    int n_read = 0;
                    while ((!erroroccurred) && (SOURCEFILELENGTH != bytecounter)) {
                        n_read = fread(inputbuffer, sizeof(char), ONE_K, fin);
                        bytecounter += (long long)n_read;
                        if (!(bytecounter % (16 * ONE_K))) {
                            if (refresh()) {
                                char lbl_val[50];
                                double p = ((double)bytecounter / (double)SOURCEFILELENGTH) * (double)100;
                                sprintf(lbl_val, "%d %%\0", (int)p);
                                progress_bar->value(p);
                                progress_bar->label(lbl_val);
                                progress_window->cursor(FL_CURSOR_WAIT);
                                Fl::check();
                            }
                        }
                        if (changeoutputfile) {
                            fclose(fout);
                            if (!strcmp(currext, "-zz")) {
                                erroroccurred = 1;
                                display_error_message(fl_xid(progress_window), "Too many output files: choose a larger size!\n(Abnormal termination; please\nremove the output files manually.)");
                                fclose(fout);
                                fclose(fin);
                                return;
                            } else {
                                nextsplitextension(currext);
                            }
                            int tmp_l = strlen(outputfilename);
                            char outputfilename_new[FL_PATH_MAX + 1];
                            strcpy(outputfilename_new, outputfilename);
                            for (int i = 1; i < 6; i++) outputfilename_new[tmp_l - i] = '\0';
                            strcat(outputfilename_new, ".s");
                            strcat(outputfilename_new, currext);
                            if ((fout = fopen(outputfilename_new, "wb")) == NULL) {
                                erroroccurred = 1;
                                display_error_message(fl_xid(progress_window), "Error creating output file!");
                            } else  outputbytecounter = 0;
                        }
                        int n_written = fwrite(inputbuffer, sizeof(char), n_read, fout);
                        if (n_written != n_read) {
                            erroroccurred = 1;
                            fclose(fout);
                            display_error_message(fl_xid(progress_window), "ERROR writing output file!!!");
                        }
                        outputbytecounter += n_read;
                        changeoutputfile = (outputbytecounter >= OUTFILESIZELIMIT) ? 1 : 0;

                        // closed window check (forced exit)
                        if (!progress_window->shown()) {
                            progress_window->show();
                            display_warning_message(
                                fl_xid(progress_window),
                                "The process has been interrupted!\nAbnormal termination; please\nremove the output files manually.");
                            fclose(fout);
                            fclose(fin);
                            return;
                        }
                    }
                    if (!erroroccurred) {
                        fclose(fout);
                        progress_bar->label("done !");
                        Fl::check();
                        display_message(fl_xid(progress_window), "Input file has been successfully split.");
                    }
                } else {
                    display_error_message(fl_xid(progress_window), "Error creating output file!");
                }
                fclose(fin);
            } else {
                display_error_message(fl_xid(progress_window), "Cannot open input file!");
            }
        }        
    }
}

void nextsplitextension(char *ext) {
    if (strlen(ext) == 3) {
        ext[1] = tolower(ext[1]);
        ext[2] = tolower(ext[2]);
        if (ext[2] < 'z') ext[2]++;
        else {
            ext[2] = 'a';
            if (ext[1] < 'z') ext[1]++;
        }
    }
}

long long readfilelength(const char *fname) {
    struct _stati64 stattest;
    int r = _stati64(fname, &stattest);
    if (r == 0) {
        __int64 len = stattest.st_size;
        return (long long) len;
    } else {
        return (long long) 0;
    }
}

inline int refresh() {
    static int seconds = -1;
    int curr_seconds;
    struct tm *ptr;
    time_t tm = time(NULL);
    ptr = localtime(&tm);
    //curr_seconds = 60 * ptr->tm_min + ptr->tm_sec;
    curr_seconds = ptr->tm_sec;
    if (abs(seconds - curr_seconds) > 1) {
        seconds = curr_seconds;
        return 1;
    }
    return 0;
}
